from itertools import zip_longest
from pathlib import Path
from typing import Callable, Generator, Iterable

from reling.asr import ASRClient
from reling.db.models import Language
from reling.helpers.input import get_input, TranscriberParams
from reling.helpers.output import output, SentenceData
from reling.tts import TTSVoiceClient
from reling.types import DialogueExchangeData
from reling.utils.transformers import get_numbering_prefix
from .types import ExchangeWithTranslation, SentenceWithTranslation

__all__ = [
    'collect_dialogue_translations',
    'collect_text_translations',
]

HIDDEN_TEXT = '(...)'
TRANSLATION_PROMPT = 'Translation: '


def collect_text_translations(
        sentences: Iterable[str],
        target_language: Language,
        source_tts: TTSVoiceClient | None,
        asr: ASRClient | None,
        hide_prompts: bool,
        storage: Path,
        on_pause: Callable[[], None],
        on_resume: Callable[[], None],
) -> Generator[SentenceWithTranslation, None, None]:
    """Collect the translations of text sentences."""
    collected: list[str] = []
    for index, sentence in enumerate(sentences):
        output(SentenceData.from_tts(
            sentence,
            source_tts,
            print_text=HIDDEN_TEXT if hide_prompts else None,
            print_prefix=get_numbering_prefix(index),
        ))
        translation = get_input(
            on_pause=on_pause,
            on_resume=on_resume,
            prompt=TRANSLATION_PROMPT,
            transcriber_params=TranscriberParams(
                transcribe=asr.get_transcriber(target_language, '\n'.join(collected)),
                storage=storage,
            ) if asr else None,
        )
        collected.append(translation.text)
        yield SentenceWithTranslation(sentence, translation)
        print()


def collect_dialogue_translations(
        exchanges: Iterable[DialogueExchangeData],
        original_translations: Iterable[DialogueExchangeData],
        target_language: Language,
        source_user_tts: TTSVoiceClient | None,
        target_speaker_tts: TTSVoiceClient | None,
        asr: ASRClient | None,
        hide_prompts: bool,
        storage: Path,
        on_pause: Callable[[], None],
        on_resume: Callable[[], None],
) -> Generator[ExchangeWithTranslation, None, None]:
    """Collect the translations of user turns in a dialogue."""
    speaker_translations: list[str] = []
    collected: list[str] = []
    for index, (exchange, original_translation) in enumerate(zip(exchanges, original_translations)):
        output(SentenceData.from_tts(
            original_translation.speaker,
            target_speaker_tts,
            print_text=HIDDEN_TEXT if hide_prompts else None,
        ))
        output(SentenceData.from_tts(
            exchange.user,
            source_user_tts,
            print_text=HIDDEN_TEXT if hide_prompts else None,
            print_prefix=get_numbering_prefix(index),
        ))
        speaker_translations.append(original_translation.speaker)
        translation = get_input(
            on_pause=on_pause,
            on_resume=on_resume,
            prompt=TRANSLATION_PROMPT,
            transcriber_params=TranscriberParams(
                transcribe=asr.get_transcriber(target_language, '\n'.join(
                    turn for exchange in zip_longest(speaker_translations, collected) for turn in exchange if turn
                )),
                storage=storage,
            ) if asr else None,
        )
        collected.append(translation.text)
        yield ExchangeWithTranslation(exchange, translation)
        print()



from datetime import datetime, timedelta

from .time import now

__all__ = [
    'TimeTracker',
]


class TimeTracker:
    started_at: datetime
    finished_at: datetime | None
    total_pause_time: timedelta
    _paused_at: datetime | None

    def __init__(self) -> None:
        self.started_at = now()
        self.finished_at = None
        self.total_pause_time = timedelta()
        self._paused_at = None

    def stop(self) -> None:
        self.finished_at = now()

    def pause(self) -> None:
        if self._paused_at is not None:
            raise RuntimeError('Already paused')
        self._paused_at = now()

    def resume(self) -> None:
        if self._paused_at is None:
            raise RuntimeError('Not paused')
        self.total_pause_time += now() - self._paused_at
        self._paused_at = None



from dataclasses import dataclass
from datetime import timedelta
from functools import partial
from itertools import repeat
from typing import Iterable

from lcs2 import diff_ranges
from rich.text import Text

from reling.app.config import MAX_SCORE
from reling.helpers.colors import default, fade, green, red
from reling.helpers.output import output, SentenceData
from reling.helpers.scoring import format_average_score
from reling.helpers.wave import play
from reling.tts import TTSVoiceClient
from reling.types import DialogueExchangeData, Input
from reling.utils.strings import tokenize
from reling.utils.time import format_time_delta
from reling.utils.transformers import get_numbering_prefix
from .types import ExchangeWithTranslation, ScoreWithSuggestion, SentenceWithTranslation

__all__ = [
    'present_dialogue_results',
    'present_text_results',
]

NOTHING_TO_IMPROVE = fade('(no changes needed)')


@dataclass
class TitleData:
    text: str
    should_number: bool
    tts: TTSVoiceClient | None


def colorize_diff(worse: str, better: str) -> tuple[Text, Text]:
    """Return the formatted pair of strings, highlighting the difference between the two."""
    tokenizer = (tokenize
                 if any(' ' in sentence for sentence in (worse, better))
                 else list)  # Tokenize into characters for, e.g., Chinese and Japanese
    worse_tokens = tokenizer(worse)
    better_tokens = tokenizer(better)
    diff = diff_ranges(*([token.lower() for token in tokens] for tokens in (worse_tokens, better_tokens)))
    worse_in_diff = {index for worse, _ in diff for index in range(worse.start, worse.stop)}
    better_in_diff = {index for _, better in diff for index in range(better.start, better.stop)}
    return (
        sum(((red if index in worse_in_diff else default)(token)
             for index, token in enumerate(worse_tokens)), Text('')),
        sum(((green if index in better_in_diff else default)(token)
             for index, token in enumerate(better_tokens)), Text('')),
    )


def format_provided_and_suggestion(
        provided: str | None,
        perfect: str | None,
) -> tuple[str | Text | None, str | Text | None]:
    """Compute the representation of the provided answer and suggested improvement to be shown to the user."""
    if provided is not None and perfect is not None:
        if provided == perfect:
            return provided, NOTHING_TO_IMPROVE
        else:
            return colorize_diff(provided, perfect)
    else:
        return provided, perfect


def present_results(
        titles: Iterable[list[TitleData]],
        provided_translations: Iterable[Input],
        original_translations: Iterable[str] | None,
        results: Iterable[ScoreWithSuggestion],
        duration: timedelta,
        target_tts: TTSVoiceClient | None,
) -> None:
    """Present the results of scoring translations."""
    scores: list[int] = []
    for index, (title_items, provided_translation, original_translation, result) in enumerate(zip(
            titles,
            provided_translations,
            original_translations if original_translations is not None else repeat(None),
            results,
    )):
        scores.append(result.score)
        for title in title_items:
            output(SentenceData.from_tts(
                title.text,
                title.tts,
                print_prefix=get_numbering_prefix(index) if title.should_number else '',
            ))
        print(f'Your score: {result.score}/{MAX_SCORE}')
        provided_text = provided_translation.text.strip() or None
        perfect_text = ((result.suggestion if result.suggestion != provided_text else None)
                        or (provided_text if result.score == MAX_SCORE else None)) if provided_text else None
        provided_print_text, improved_print_text = format_provided_and_suggestion(provided_text, perfect_text)
        output(*[
            SentenceData(
                print_text=provided_print_text,
                print_prefix='Provided: ',
                reader=partial(play, provided_translation.audio) if provided_translation.audio and target_tts else None,
                reader_id='provided',
            ),
            SentenceData.from_tts(
                text=perfect_text,
                client=target_tts,
                print_text=improved_print_text,
                print_prefix='Improved: ',
                reader_id='improved',
            ),
            *([SentenceData.from_tts(
                text=original_translation,
                client=target_tts,
                print_prefix='Original: ',
                reader_id='original',
            )] if original_translation is not None else []),
        ])
        print()
    print(f'Average score: {format_average_score(scores)}')
    print(f'Exam duration: {format_time_delta(duration)}')


def present_text_results(
        sentences: Iterable[SentenceWithTranslation],
        original_translations: Iterable[str],
        show_original: bool,
        results: Iterable[ScoreWithSuggestion],
        duration: timedelta,
        source_tts: TTSVoiceClient | None,
        target_tts: TTSVoiceClient | None,
) -> None:
    """Present the results of scoring text translations."""
    present_results(
        titles=([
            TitleData(
                text=sentence.sentence,
                should_number=True,
                tts=source_tts,
            ),
        ] for sentence in sentences),
        provided_translations=(sentence.translation for sentence in sentences),
        original_translations=original_translations if show_original else None,
        results=results,
        duration=duration,
        target_tts=target_tts,
    )


def present_dialogue_results(
        exchanges: Iterable[ExchangeWithTranslation],
        original_translations: Iterable[DialogueExchangeData],
        show_original: bool,
        results: Iterable[ScoreWithSuggestion],
        duration: timedelta,
        source_user_tts: TTSVoiceClient | None,
        target_speaker_tts: TTSVoiceClient | None,
        target_user_tts: TTSVoiceClient | None,
) -> None:
    """Present the results of scoring dialogue translations."""
    present_results(
        titles=([
            TitleData(
                text=original_translation.speaker,
                should_number=False,
                tts=target_speaker_tts,
            ),
            TitleData(
                text=exchange.exchange.user,
                should_number=True,
                tts=source_user_tts,
            ),
        ] for original_translation, exchange in zip(
            original_translations,
            exchanges,
        )),
        provided_translations=(exchange.user_translation for exchange in exchanges),
        original_translations=(exchange.user for exchange in original_translations) if show_original else None,
        results=results,
        duration=duration,
        target_tts=target_user_tts,
    )



from itertools import starmap
from typing import Generator

from reling.db.enums import Gender
from reling.db.models import Language
from reling.gpt import GPTClient
from reling.types import DialogueExchangeData
from reling.utils.iterables import pair_items
from reling.utils.transformers import add_numbering, apply, omit_empty, remove_numbering, strip

__all__ = [
    'translate_dialogue_exchanges',
    'translate_text_sentences',
]


def translate_text_sentences(
        gpt: GPTClient,
        sentences: list[str],
        source_language: Language,
        target_language: Language,
) -> Generator[str, None, None]:
    return gpt.ask(
        '\n'.join([
            f'Translate the following text from {source_language.name} into {target_language.name}.',
            f'Generate only the specified translations without any additional text.',
            f'Number each translated sentence and place each on a new line.',
            f'---',
            *apply(add_numbering, sentences),
        ]),
        transformers=[strip, omit_empty, remove_numbering],
    )


def translate_dialogue_exchanges(
        gpt: GPTClient,
        exchanges: list[DialogueExchangeData],
        speaker_gender: Gender,
        user_gender: Gender,
        source_language: Language,
        target_language: Language,
) -> Generator[DialogueExchangeData, None, None]:
    DialogueExchangeData.assert_speaker_comes_first()
    return iter(starmap(DialogueExchangeData, pair_items(gpt.ask(
        '\n'.join([
            f'Translate the following dialogue between {speaker_gender.describe()} and {user_gender.describe()} '
            f'from {source_language.name} into {target_language.name}'
            f'{f' ({speaker_gender.describe()} speaks in the odd-numbered sentences,'
               f' and {user_gender.describe()} responds in the even-numbered sentences)'
               if speaker_gender != user_gender
               else ''}'
            f'.',

            f'Generate only the specified translations without any additional text.',
            f'Number each translated sentence and place each on a new line.',
            f'---',
            *apply(add_numbering, [turn for exchange in exchanges for turn in exchange.all()]),
        ]),
        transformers=[strip, omit_empty, remove_numbering],
    ))))



from pathlib import Path
import re

from openai import OpenAI

from reling.db.models import Language
from reling.helpers.openai import openai_handler
from reling.types import Transcriber
from reling.utils.strings import capitalize_first_char, universal_normalize

__all__ = [
    'ASRClient',
]


class ASRClient:
    _client: OpenAI
    _model: str

    def __init__(self, *, api_key: str, model: str) -> None:
        self._client = OpenAI(api_key=api_key)
        self._model = model

    @staticmethod
    def _normalize_transcription(text: str) -> str:
        return capitalize_first_char(re.sub(
            r'^[a-zA-Z\u00C0-\u024F¹²³⁴-⁹]+(?=[A-Z][a-z\'’]+([.?!,;:]|\s))',
            '',  # ^ Remove extra prefixes that Whisper sometimes adds
            ''.join(char for char in text.strip()
                    if not (0xE000 <= ord(char) <= 0xF8FF))  # Strip private use area characters from the Whisper output
        ))

    def transcribe(self, file: Path, language: Language | None = None, context: str | None = None) -> str:
        """Transcribe an audio file."""
        with openai_handler():
            return universal_normalize(self._normalize_transcription(self._client.audio.transcriptions.create(
                file=file.open('rb'),
                model=self._model,
                language=language.short_code if language else None,
                prompt=context,
            ).text))

    def get_transcriber(self, language: Language | None = None, context: str | None = None) -> Transcriber:
        def transcribe(file: Path) -> str:
            return self.transcribe(file, language, context)
        return transcribe



from rich.text import Text

__all__ = [
    'default',
    'fade',
    'green',
    'red',
]

FADE = 'grey50'
RED = 'red'
GREEN = 'green'
DEFAULT = 'default'


def fade(text: str) -> Text:
    """Return the text in a faded color."""
    return Text(text, style=FADE)


def red(text: str) -> Text:
    """Return the text in red."""
    return Text(text, style=RED)


def green(text: str) -> Text:
    """Return the text in green."""
    return Text(text, style=GREEN)


def default(text: str) -> Text:
    """Return the text in the default color."""
    return Text(text, style=DEFAULT)



from contextlib import contextmanager
from sqlite3 import Connection
from typing import Any, Generator

from sqlalchemy import create_engine, event, Engine
from sqlalchemy.orm import Session

from . import models  # Register models with SQLAlchemy
from .base import Base
from .migrations import migrate

__all__ = [
    'init_db',
    'migrate',
    'Session',
    'single_session',
]

SESSION: Session | None = None


@event.listens_for(Engine, 'connect')
def set_sqlite_pragma(dbapi_connection: Connection, _connection_record: Any) -> None:
    """
    Enable foreign key support for SQLite.
    See https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support
    """
    cursor = dbapi_connection.cursor()
    cursor.execute('PRAGMA foreign_keys=ON')
    cursor.close()


def init_db(url: str) -> None:
    global SESSION
    if SESSION is None:
        engine = create_engine(url)
        Base.metadata.create_all(engine)
        SESSION = Session(engine)


@contextmanager
def single_session() -> Generator[Session, None, None]:
    if SESSION is None:
        raise RuntimeError('Database is not initialized')
    try:
        yield SESSION
    except Exception:
        SESSION.rollback()
        raise



from .archive import archive
from .create import create
from .db import db
from .delete import delete
from .exam import exam
from .history import history
from .list import list_
from .rename import rename
from .show import show
from .stats import stats
from .unarchive import unarchive

__all__ = [
    'archive',
    'create',
    'db',
    'delete',
    'exam',
    'history',
    'list_',
    'show',
    'stats',
    'unarchive',
]



from pathlib import Path
import sqlite3

__all__ = [
    'migrate',
]

ZERO_DATETIME = "'1970-01-01 00:00:00'"


def get_migration_commands(from_version: str) -> list[str]:
    match from_version:
        case 'a':
            return [
                f'ALTER TABLE text_exams add total_pause_time DATETIME NOT NULL default {ZERO_DATETIME}',
                f'ALTER TABLE dialogue_exams add total_pause_time DATETIME NOT NULL default {ZERO_DATETIME}',
            ]
        case _:
            raise ValueError(f'Unknown migration from version {from_version}')


def migrate(database: Path, from_version: str) -> None:
    connection = sqlite3.connect(database)
    cursor = connection.cursor()
    for command in get_migration_commands(from_version):
        cursor.execute(command)
    connection.commit()
    connection.close()



from contextlib import contextmanager
from typing import Generator

from .strings import universal_normalize

__all__ = [
    'clear_current_line',
    'clear_previous',
    'erase_previous',
    'input_and_erase',
    'interruptible_input',
    'print_and_erase',
]


def clear_current_line() -> None:
    print('\033[2K', end='\r')


def clear_previous(lines: int = 1) -> None:
    print('\033[F\033[K' * lines, end='\r')


def erase_previous(text: str, include_extra_line: bool = True) -> None:
    clear_current_line()
    clear_previous(text.count('\n') + (1 if include_extra_line else 0))


def interruptible_input(prompt: str) -> str:
    try:
        return universal_normalize(input(prompt))
    except KeyboardInterrupt:
        erase_previous(prompt, include_extra_line=False)
        raise


def input_and_erase(prompt: str) -> str:
    data = interruptible_input(prompt)
    erase_previous(prompt)
    return data


@contextmanager
def print_and_erase(text: str) -> Generator[None, None, None]:
    print(text)
    yield
    erase_previous(text)



import re
from typing import Callable, Iterable, Generator

from .strings import universal_normalize

__all__ = [
    'add_numbering',
    'apply',
    'get_numbering_prefix',
    'normalize',
    'omit_empty',
    'remove_numbering',
    'slugify',
    'strip',
    'Transformer',
]

type Transformer = Callable[[str, int], str | None]
# The second argument is the index of the item in the list.


def apply(transformer: Transformer, items: Iterable[str]) -> Generator[str, None, None]:
    for index, item in enumerate(items):
        yield transformer(item, index)


def get_numbering_prefix(index: int) -> str:
    return f'{index + 1}. '


def add_numbering(text: str, index: int) -> str:
    return f'{get_numbering_prefix(index)}{text}'


def remove_numbering(text: str, _: int) -> str:
    return re.sub(r'^\s*\d+[.)]\s+', '', text)


def strip(text: str, _: int) -> str:
    return text.strip()


def omit_empty(text: str, _: int) -> str | None:
    return text or None


def slugify(text: str, _: int) -> str:
    return re.sub(r'[^\w-]', '', re.sub(r'\s+', '-', text.lower().strip()))


def normalize(text: str, _: int) -> str:
    return universal_normalize(text)



from datetime import datetime
import sys
from typing import cast

from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import CONTENT_ARG, LANGUAGE_OPT, LANGUAGE_OPT_FROM
from reling.db.models import DialogueExam, Language, TextExam
from reling.helpers.scoring import format_average_score
from reling.utils.tables import build_table, print_table
from reling.utils.time import format_time, format_time_delta

__all__ = [
    'history',
]

FROM = 'From'
TO = 'To'
TAKEN_AT = 'Taken at'
DURATION = 'Duration'
SCORE = 'Score'

AUDIO_OUTPUT = '🔈'
AUDIO_INPUT = '🎤'


def match(exam: TextExam | DialogueExam, from_: Language | None, to: Language | None) -> bool:
    return ((from_ is None or cast(Language, exam.source_language).id == from_.id)
            and (to is None or cast(Language, exam.target_language).id == to.id))


def get_sort_key(exam: TextExam | DialogueExam) -> tuple:
    return (
        cast(Language, exam.source_language).name,
        cast(Language, exam.target_language).name,
        -cast(datetime, exam.started_at).timestamp(),
    )


@app.command()
def history(content: CONTENT_ARG, from_: LANGUAGE_OPT_FROM = None, to: LANGUAGE_OPT = None) -> None:
    """Display exam history, optionally filtered by source or target language."""
    set_default_content(content)
    exams = sorted(
        filter(lambda e: match(e, from_, to), content.exams),
        key=get_sort_key,
    )
    if exams:
        table = build_table(
            headers=[
                FROM,
                TO,
                TAKEN_AT,
                DURATION,
                SCORE,
            ],
            justify={
                FROM: 'left',
                TO: 'left',
                TAKEN_AT: 'left',
                DURATION: 'right',
                SCORE: 'right',
            },
            data=[{
                FROM: ' '.join([
                    cast(Language, exam.source_language).name,
                    *([AUDIO_OUTPUT] if exam.read_source else []),
                ]),
                TO: ' '.join([
                    cast(Language, exam.target_language).name,
                    *([AUDIO_OUTPUT] if exam.read_target else []),
                    *([AUDIO_INPUT] if exam.listened else []),
                ]),
                TAKEN_AT: format_time(cast(datetime, exam.started_at)),
                DURATION: format_time_delta(exam.duration),
                SCORE: format_average_score([result.score for result in exam.results]),
            } for exam in exams],
            group_by=[
                FROM,
                TO,
            ],
        )
        print_table(table)
    else:
        print('No exams found.', file=sys.stderr)



from dataclasses import dataclass

from reling.types import DialogueExchangeData, Input

__all__ = [
    'ExchangeWithTranslation',
    'PreScoreWithSuggestion',
    'ScoreWithSuggestion',
    'SentenceWithTranslation',
]


@dataclass
class SentenceWithTranslation:
    sentence: str
    translation: Input


@dataclass
class ExchangeWithTranslation:
    exchange: DialogueExchangeData
    user_translation: Input


@dataclass
class PreScoreWithSuggestion:
    score: int
    suggestion: str | None


@dataclass
class ScoreWithSuggestion:
    score: int
    suggestion: str | None



from typing import Generator

from openai import OpenAI

from reling.helpers.openai import openai_handler
from reling.utils.feeders import Feeder, LineFeeder
from reling.utils.transformers import normalize, Transformer

__all__ = [
    'GPTClient',
]

CREATIVE_TEMPERATURE = 1.0


class GPTClient:
    _client: OpenAI
    _model: str

    def __init__(self, *, api_key: str, model: str) -> None:
        self._client = OpenAI(api_key=api_key)
        self._model = model

    def ask(
            self,
            prompt: str,
            creative: bool = True,
            feeder_type: type[Feeder] = LineFeeder,
            transformers: list[Transformer] | None = None,
            auto_normalize: bool = True,
    ) -> Generator[str, None, None]:
        """
        Ask the model a question and yield sections of the response as they become available, applying transformers.
        """
        feeder = feeder_type()

        with openai_handler():
            stream = self._client.chat.completions.create(
                model=self._model,
                stream=True,
                messages=[{'role': 'user', 'content': prompt}],
                temperature=CREATIVE_TEMPERATURE if creative else 0.0,
            )

        section_index = 0

        def flush() -> Generator[str, None, None]:
            nonlocal section_index
            while (section := feeder.get()) is not None:
                for transformer in (transformers or []) + ([normalize] if auto_normalize else []):
                    section = transformer(section, section_index)
                    if section is None:
                        break
                else:
                    yield section
                    section_index += 1

        for chunk in stream:
            feeder.put(chunk.choices[0].delta.content or '')
            yield from flush()

        feeder.end()
        yield from flush()



from math import floor
from statistics import mean

from lcs2 import diff, lcs_length

from reling.app.config import MAX_SCORE
from reling.utils.strings import tokenize

__all__ = [
    'calculate_diff_score',
    'format_average_score',
]


def format_average_score(scores: list[int]) -> str:
    return f'{mean(scores):.1f}'


def calculate_char_diff_score(a: str, b: str) -> int:
    """Return the score based on the longest common subsequence of two strings."""
    if a == b:  # Avoid division by zero
        return MAX_SCORE
    return floor(lcs_length(a, b) / max(len(a), len(b)) * MAX_SCORE)


def calculate_mistake_score(mistakes: int) -> int:
    """Return the score based on the number of mistakes."""
    return floor(MAX_SCORE * ((1 - 1 / MAX_SCORE) ** mistakes))


def calculate_word_diff_score(a: str, b: str) -> int:
    """Return the score based on the diff between two strings tokenized into lowercase words."""
    a_words, b_words = ([word.lower() for word in tokenize(sentence, words_only=True)] for sentence in (a, b))
    a_diff: list[str] = []
    b_diff: list[str] = []
    mistakes = 0
    for a_tokens, b_tokens in diff(a_words, b_words):
        mistakes += max(len(a_tokens), len(b_tokens))
        a_diff.extend(a_tokens)
        b_diff.extend(b_tokens)
    mistakes -= lcs_length(a_diff, b_diff)
    return calculate_mistake_score(mistakes)


def calculate_diff_score(a: str, b: str) -> int:
    """Return the score based on the diff between two strings."""
    return min(
        calculate_char_diff_score(a, b),
        calculate_word_diff_score(a, b),
    )



from pathlib import Path
from typing import cast

from sqlalchemy import ColumnElement
from sqlalchemy.sql import exists, func

from reling.db import single_session
from reling.db.models import Speaker, Style, Topic
from reling.utils.csv import read_csv

__all__ = [
    'get_random_modifier',
    'populate_modifiers',
]


def populate_modifiers[T: (Speaker, Style, Topic)](entity_class: type[T], data: Path) -> None:
    """Populate the table with data from the CSV file, if the table is empty."""
    with single_session() as session:
        if not session.query(exists().where(cast(ColumnElement[str], entity_class.name).is_not(None))).scalar():
            for entity in read_csv(data, ['name']):
                session.add(entity_class(**entity))
        session.commit()


def get_random_modifier[T: (Speaker, Style, Topic)](entity_class: type[T]) -> T:
    """Get a random entity from the given table."""
    with single_session() as session:
        entity = session.query(entity_class).order_by(func.random()).first()
    if entity is None:
        raise RuntimeError(f'Table "{entity_class.__tablename__}" is empty.')
    return entity
